@@ -1,36 +1,28 @@
-
-import time
 import h5py
 import glob
 import json
-import requests
 import mdf_toolbox
 from json2table import convert
+import numpy as np
 import pandas as pd
-from datetime import date
 from typing import Any
-import multiprocessing
+import logging
+import warnings
+import os
+
 from mdf_connect_client import MDFConnectClient
 from mdf_forge import Forge
-# from dlhub_sdk.utils.schemas import validate_against_dlhub_schema
-# from dlhub_sdk.models.servables.keras import KerasModel
-# from dlhub_sdk.models.servables.sklearn import ScikitLearnModel
 from dlhub_sdk import DLHubClient
-from collections import namedtuple
-from joblib import Parallel, delayed
-from pydantic import AnyUrl, ValidationError
+
 from foundry.models import (
     FoundryMetadata,
     FoundryConfig,
-    FoundrySpecificationDataset,
-    FoundrySpecification,
     FoundryDataset
 )
-import logging
-import os
-import shutil
-from collections import deque
-logging.disable(logging.INFO)
+from foundry.https_download import download_file, recursive_ls
+
+
+logger = logging.getLogger(__name__)
 
 
 class Foundry(FoundryMetadata):
@@ -110,7 +102,7 @@ class Foundry(FoundryMetadata):
             data_mdf_authorizer=auths["data_mdf"],
             petrel_authorizer=auths["petrel"],
         )
-        
+
         self.transfer_client = auths['transfer']
 
         if index == "mdf":
@@ -123,7 +115,7 @@ class Foundry(FoundryMetadata):
             authorizer=auths["mdf_connect"], test=test
         )
 
-        ## TODO: come back to add in DLHub functionality after globus-sdk>=3.0 supported
+        # TODO: come back to add in DLHub functionality after globus-sdk>=3.0 supported
         self.dlhub_client = DLHubClient(
             dlh_authorizer=auths["dlhub"],
             search_authorizer=auths["search_authorizer"],
@@ -134,7 +126,6 @@ class Foundry(FoundryMetadata):
             force_login=False,
         )
 
-
         self.xtract_tokens = {
             "auth_token": auths["petrel"].access_token,
             "transfer_token": auths["transfer"].authorizer.access_token,
@@ -167,9 +158,6 @@ class Foundry(FoundryMetadata):
         if metadata:
             res = metadata
 
-        if metadata:
-            res = metadata
-
         # MDF specific logic
         if is_doi(name) and not metadata:
             res = self.forge_client.match_resource_types("dataset")
@@ -182,11 +170,13 @@ class Foundry(FoundryMetadata):
             res = res.match_field("mdf.source_id", name).search()
 
         # unpack res, handle if empty
-        try:
-            # if search returns multiple results, this automatically uses first result
-            res = res[0]
-        except IndexError as e:
-            raise Exception("load: No metadata found for given dataset") from e
+        if len(res) == 0:
+            raise Exception(f"load: No metadata found for given dataset {name}")
+
+        # if search returns multiple results, this automatically uses first result, while warning the user
+        if len(res) > 1:
+            warnings.warn("Multiple datasets found for the given search query. Using first dataset")
+        res = res[0]
 
         try:
             res["dataset"] = res["projects"][self.config.metadata_key]
@@ -198,48 +188,58 @@ class Foundry(FoundryMetadata):
         # TODO: Creating a new Foundry instance is a problematic way to update the metadata,
         # we should find a way to abstract this.
 
-        fdataset = FoundryDataset(**res['dataset'])
         self.dc = res['dc']
         self.mdf = res['mdf']
-        self.dataset = fdataset
+        self.dataset = FoundryDataset(**res['dataset'])
 
-
-        if download is True:  # Add check for package existence
+        if download:  # Add check for package existence
             self.download(
                 interval=kwargs.get("interval", 10), globus=globus, verbose=verbose
             )
 
         return self
 
-    def list(self):
-        """List available Foundry data packages
+    def search(self, q=None, limit=None):
+        """Search available Foundry datasets
+        q (str): query string to match
+        limit (int): maximum number of results to return
 
         Returns
         -------
             (pandas.DataFrame): DataFrame with summary list of Foundry data packages including name, title, publication year, and DOI
         """
+        if not q:
+            q = None
         res = (
             self.forge_client.match_field(
                 "mdf.organizations", self.config.organization)
             .match_resource_types("dataset")
-            .search()
+            .search(q, limit=limit)
         )
 
-
         return pd.DataFrame(
             [
                 {
                     "source_id": r["mdf"]["source_id"],
                     "name": r["dc"]["titles"][0]["title"],
                     "year": r["dc"].get("publicationYear", None),
-                    "DOI": r["dc"]["identifier"]["identifier"], 
+                    "DOI": r["dc"].get("identifier", {}).get("identifier", None),
                 }
                 for r in res
             ]
         )
 
+    def list(self):
+        """List available Foundry datasets
+
+        Returns
+        -------
+            (pandas.DataFrame): DataFrame with summary list of Foundry datasets including name, title, publication year, and DOI
+        """
+        return self.search()
+
     def get_packages(self, paths=False):
-        """Get available local data packages
+        """Get available local datasets
 
         Args:
            paths (bool): If True return paths in addition to package, if False return package name only
@@ -267,7 +267,6 @@ class Foundry(FoundryMetadata):
         """
         if not packages:
             packages = self.get_packages()
-        f = Foundry()
 
         X_frames = []
         y_frames = []
@@ -282,20 +281,23 @@ class Foundry(FoundryMetadata):
 
         return pd.concat(X_frames), pd.concat(y_frames)
 
-    def run(self, name, inputs, **kwargs):
+    def run(self, name, inputs, funcx_endpoint=None, **kwargs):
         """Run a model on data
 
         Args:
            name (str): DLHub model name
            inputs: Data to send to DLHub as inputs (should be JSON serializable)
+           funcx_endpoint (optional): UUID for the funcx endpoint to run the model on, if not the default (eg River)
 
         Returns
         -------
              Returns results after invocation via the DLHub service
         """
+        if funcx_endpoint is not None:
+            self.dlhub_client.fx_endpoint = funcx_endpoint
         return self.dlhub_client.run(name, inputs=inputs, **kwargs)
 
-    def load_data(self, source_id=None, as_hdf5=False):
+    def load_data(self, source_id=None, globus=True, as_hdf5=False):
         """Load in the data associated with the prescribed dataset
 
         Tabular Data Type: Data are arranged in a standard data frame
@@ -323,10 +325,10 @@ class Foundry(FoundryMetadata):
             if self.dataset.splits:
                 for split in self.dataset.splits:
                     data[split.label] = self._load_data(file=split.path,
-                                                        source_id=source_id, as_hdf5=as_hdf5)
+                                                        source_id=source_id, globus=globus, as_hdf5=as_hdf5)
                 return data
             else:
-                return {"data": self._load_data(source_id=source_id, as_hdf5=as_hdf5)}
+                return {"data": self._load_data(source_id=source_id, globus=globus, as_hdf5=as_hdf5)}
         except Exception as e:
             raise Exception(
                 "Metadata not loaded into Foundry object, make sure to call load()") from e
@@ -344,8 +346,6 @@ class Foundry(FoundryMetadata):
             buf = f'<h2>{title}</h2>{authors}<p>{DOI}</p>'
 
             buf = f'{buf}<h3>Dataset</h3>{convert(json.loads(self.dataset.json(exclude={"dataframe"})))}'
-        # buf = f'{buf}<h3>MDF</h3>{convert(self.mdf)}'
-        # buf = f'{buf}<h3>DataCite</h3>{convert(self.dc)}'
         return buf
 
     def publish(self, foundry_metadata, data_source, title, authors, update=False,
@@ -596,17 +596,17 @@ class Foundry(FoundryMetadata):
                         split.path = split.path[1:]
                     if not os.path.isfile(os.path.join(path, split.path)):
                         missing_files.append(split.path)
-                #if number of missing files is greater than zero, redownload with informative message
+                # if number of missing files is greater than zero, redownload with informative message
                 if len(missing_files) > 0:
-                    print(f"Dataset will be redownloaded, following files are missing: {missing_files}")
+                    logger.info(f"Dataset will be redownloaded, following files are missing: {missing_files}")
                 else:
                     return self
             else:
                 # in the case of no splits, ensure the directory contains at least one file
-                if(len(os.listdir(path)) >= 1):
+                if (len(os.listdir(path)) >= 1):
                     return self
                 else:
-                    print("Dataset will be redownloaded, expected file is missing")
+                    logger.info("Dataset will be redownloaded, expected file is missing")
 
         res = self.forge_client.search(
             f"mdf.source_id:{self.mdf['source_id']}", advanced=True
@@ -620,17 +620,15 @@ class Foundry(FoundryMetadata):
                 download_datasets=True,
             )
         else:
-
-            source_id = self.mdf["source_id"]
             https_config = {
                 "source_ep_id": "82f1b5c6-6e9b-11e5-ba47-22000b92c6ec",
                 "base_url": "https://data.materialsdatafacility.org",
                 "folder_to_crawl": f"/foundry/{self.mdf['source_id']}/",
-                "source_id":self.mdf["source_id"]
+                "source_id": self.mdf["source_id"]
             }
-            
-            task_list = list(recursive_ls(self.transfer_client, 
-                                          https_config['source_ep_id'], 
+
+            task_list = list(recursive_ls(self.transfer_client,
+                                          https_config['source_ep_id'],
                                           https_config['folder_to_crawl']))
             # TODO Add parallel
             for task in task_list:
@@ -642,7 +640,7 @@ class Foundry(FoundryMetadata):
             if self.dataset.splits:
                 missing_files = []
                 for split in self.dataset.splits:
-                    if split.path[0] == '/': # if absolute path, make it a relative path
+                    if split.path[0] == '/':  # if absolute path, make it a relative path
                         split.path = split.path[1:]
                     if not os.path.isfile(os.path.join(path, split.path)):
                         # keeping track of all files not downloaded
@@ -651,58 +649,13 @@ class Foundry(FoundryMetadata):
                     raise FileNotFoundError(f"Downloaded directory does not contain the following files: {missing_files}")
 
             else:
-                if(len(os.listdir(path)) < 1):
+                if (len(os.listdir(path)) < 1):
                     raise FileNotFoundError("Downloaded directory does not contain the expected file")
         else:
             raise NotADirectoryError("Unable to create directory to download data")
 
         return self
 
-    def build(self, spec, globus=False, interval=3, file=False, as_object=False):
-        """Build a Foundry Data Package
-        Args:
-            spec (multiple): dict or str (relative filename) of the data package specification
-            globus (bool): if True use Globus to fetch datasets
-            interval (int): Polling interval on checking task status in seconds.
-            type (str): One of "file" or None
-
-        Returns
-        -------
-        (Foundry): self: for chaining
-        """
-        if as_object:
-            return [key for key in self.dataset.keys if key.type == type]
-        else:
-            keys = [key.key for key in self.dataset.keys if key.type == type]
-            key_list = []
-            for k in keys:
-                key_list = key_list + k
-            return key_list
-
-        print("Building Data Package")
-        num_cores = multiprocessing.cpu_count()
-
-        def start_download(ds, interval=interval, globus=False):
-            print(f"=== Fetching Data Package {ds.name} ===")
-            f = Foundry().load(ds.name, download=False)
-            f = f.download(interval=interval, globus=globus)
-            return {"success": True}
-
-        if file:
-            with open(file, "r") as fp:
-                fs = FoundrySpecification(**json.load(fp))
-        else:
-            fs = FoundrySpecification(**spec)
-
-        fs.remove_duplicate_dependencies()
-
-        results = Parallel(n_jobs=num_cores)(
-            delayed(start_download)(ds, interval=interval, globus=globus)
-            for ds in fs.dependencies
-        )
-
-        return self
-
     def get_keys(self, type=None, as_object=False):
         """Get keys for a Foundry dataset
 
@@ -733,9 +686,7 @@ class Foundry(FoundryMetadata):
                 key_list = key_list + k
             return key_list
 
-
-    def _load_data(self, file=None, source_id=None, as_hdf5=False):
-
+    def _load_data(self, file=None, source_id=None, globus=True, as_hdf5=False):
         # Build the path to access the cached data
         if source_id:
             path = os.path.join(self.config.local_cache_dir, source_id)
@@ -743,23 +694,20 @@ class Foundry(FoundryMetadata):
             path = os.path.join(self.config.local_cache_dir,
                                 self.mdf["source_id"])
 
+        # Determine which file to load, defaults to config.dataframe_file
+        if not file:
+            file = self.config.dataframe_file
+        if path is None:
+            raise ValueError(f"Path to data file is invalid; check that dataset source_id is valid: "
+                             f"{source_id or self.mdf['source_id']}")
+        path_to_file = os.path.join(path, file)
+
+        # Check to see whether file exists at path
+        if not os.path.isfile(path_to_file):
+            raise FileNotFoundError(f"No file found at expected path: {path_to_file}")
+
         # Handle Foundry-defined types.
         if self.dataset.data_type.value == "tabular":
-            # Determine which file to load, defaults to config.dataframe_file
-            if not file:
-                file = self.config.dataframe_file
-
-
-            # Check to make sure the path can be created
-            try:
-                path_to_file = os.path.join(path, file)
-            except Exception as e:
-                print(f"Unable to find path to file for download: {e}")
-                raise e
-
-            # Check to see whether file exists at path
-            if not os.path.isfile(path_to_file):
-                raise FileNotFoundError(f"No file found at expected path: {path_to_file}")
             # If the file is not local, fetch the contents with Globus
             # Check if the contents are local
             # TODO: Add hashes and versioning to metadata and checking to the file
@@ -768,21 +716,21 @@ class Foundry(FoundryMetadata):
                     path_to_file
                 )
             except Exception as e:
-                print(f"Reading {file} as JSON failed: {e} \n", "Now attempting to read as JSONL")
+                logger.info(f"Cannot read {file} as JSON: {e} \n", "Now attempting to read as JSONL")
                 try:
                     # Try to read individual lines instead
                     self.dataset.dataframe = pd.read_json(
                         path_to_file, lines=True
                     )
                 except Exception as f:
-                    print(f"Reading {file} as JSONL failed: {f} \n", "Now attempting to read as CSV")
+                    logger.info(f"Cannot read {file} as JSONL: {f} \n", "Now attempting to read as CSV")
                     try:
-                        #Try to read as CSV instead
+                        # Try to read as CSV instead
                         self.dataset.dataframe = pd.read_csv(
                             path_to_file
                         )
                     except Exception as g:
-                        print(f"Reading {file} as CSV failed, unable to load data properly: {g}")
+                        logger.fatal(f"Cannot read {file} as CSV, failed to load data properly: {g}")
                         raise e
 
             return (
@@ -791,11 +739,7 @@ class Foundry(FoundryMetadata):
             )
 
         elif self.dataset.data_type.value == "hdf5":
-            if not file:
-                file = self.config.data_file
-
-            filepath = os.path.join(path, file)
-            f = h5py.File(filepath, "r")
+            f = h5py.File(path_to_file, "r")
             special_types = ["input", "target"]
             tmp_data = {s: {} for s in special_types}
             for s in special_types:
@@ -804,7 +748,7 @@ class Foundry(FoundryMetadata):
                         tmp_data[s][key] = f[key]
                     elif isinstance(f[key], h5py.Group):
                         if is_pandas_pytable(f[key]):
-                            df = pd.read_hdf(filepath, key)
+                            df = pd.read_hdf(path_to_file, key)
                             tmp_data[s][key] = df
                         else:
                             tmp_data[s][key] = f[key]
@@ -814,6 +758,85 @@ class Foundry(FoundryMetadata):
         else:
             raise NotImplementedError
 
+    def _get_inputs_targets(self, split: str = None):
+        """Get Inputs and Outputs from a Foundry Dataset
+
+        Arguments:
+            split (string): Split to get inputs and outputs from.
+                    **Default:** ``None``
+
+        Returns: (Tuple) Tuple of the inputs and outputs
+        """
+        raw = self.load_data(as_hdf5=False)
+
+        if not split:
+            split = self.dataset.splits[0].type
+
+        if self.dataset.data_type.value == "hdf5":
+            inputs = []
+            targets = []
+            for key in self.dataset.keys:
+                # raw[split][key.type][key.key[0]] gets the data values for the given key.
+                #
+                # For example, if the key was coordinates and had type target, then
+                # raw[split][key.type][key.key[0]] would return all the coordinates for each item
+                # and raw[split][key.type][key.key[0]].keys() are the indexes of the item.
+                if len(raw[split][key.type][key.key[0]].keys()) != self.dataset.n_items:
+                    continue
+
+                # Get a numpy array of all the values for each item for that key
+                val = np.array([raw[split][key.type][key.key[0]][k] for k in raw[split][key.type][key.key[0]].keys()])
+                if key.type == 'input':
+                    inputs.append(val)
+                else:
+                    targets.append(val)
+
+            return (inputs, targets)
+
+        elif self.dataset.data_type.value == "tabular":
+            inputs = []
+            targets = []
+
+            for index, arr in enumerate([inputs, targets]):
+                df = raw[split][index]
+                for key in df.keys():
+                    arr.append(df[key].values)
+
+            return (inputs, targets)
+
+        else:
+            raise NotImplementedError
+
+    def to_torch(self, split: str = None):
+        """Convert Foundry Dataset to a PyTorch Dataset
+
+        Arguments:
+            split (string): Split to create PyTorch Dataset on.
+                    **Default:** ``None``
+
+        Returns: (TorchDataset) PyTorch Dataset of all the data from the specified split
+
+        """
+        from foundry.loaders.torch_wrapper import TorchDataset
+
+        inputs, targets = self._get_inputs_targets(split)
+        return TorchDataset(inputs, targets)
+
+    def to_tensorflow(self, split: str = None):
+        """Convert Foundry Dataset to a Tensorflow Sequence
+
+        Arguments:
+            split (string): Split to create Tensorflow Sequence on.
+                    **Default:** ``None``
+
+        Returns: (TensorflowSequence) Tensorflow Sequence of all the data from the specified split
+
+        """
+        from foundry.loaders.tf_wrapper import TensorflowSequence
+
+        inputs, targets = self._get_inputs_targets(split)
+        return TensorflowSequence(inputs, targets)
+
 
 def is_pandas_pytable(group):
     if 'axis0' in group.keys() and 'axis1' in group.keys():
@@ -827,53 +850,3 @@ def is_doi(string: str):
         return True
     else:
         return False
-def _get_files(tc, ep, queue, max_depth):
-    while queue:
-        abs_path, rel_path, depth = queue.pop()
-        path_prefix = rel_path + "/" if rel_path else ""
-
-        res = tc.operation_ls(ep, path=abs_path)
-
-        if depth < max_depth:
-            queue.extend(
-                (
-                    res["path"] + item["name"],
-                    path_prefix + item["name"],
-                    depth + 1,
-                )
-                for item in res["DATA"]
-                if item["type"] == "dir"
-            )
-        for item in res["DATA"]:
-            if item["type"]=='file':
-                item["name"] = path_prefix + item["name"]
-                item["path"] = abs_path.replace('/~/','/')
-                yield item
-
-def recursive_ls(tc, ep, path, max_depth=3):
-    queue = deque()
-    queue.append((path, "", 0))
-    yield from _get_files(tc, ep, queue, max_depth)
-
-def download_file(item, https_config):
-        url = f"{https_config['base_url']}{item['path']}{item['name']}"
-        
-        # removes data source (eg MDF) parent directories, leaving the split path only
-        datasplit_subpath = item["path"].replace("/foundry/","/")
-
-        # build destination path for data file
-        destination = os.path.join("data/", https_config['source_id'], item['name'])
-
-        parent_path = os.path.split(destination)[0]
-
-        # if parent directories don't exist, create them
-        if not os.path.exists(parent_path):
-            os.makedirs(parent_path)
-
-        response = requests.get(url)
-
-        # write file to local destination
-        with open(destination, "wb") as f:
-            f.write(response.content)
-
-        return {destination + " status": True}
\ No newline at end of file
